from dataclasses import dataclass
from typing import List

from pyaniml.core.enums import SIUnits
from pyaniml.utility.utils import SchemaBase, element, attribute


@dataclass
class SIUnit(SchemaBase):
    """Container for a SI base unit."""

    si_name: SIUnits
    factor: float = attribute(name="factor", default_value=1)
    exponent: float = attribute(name="exponent", default_value=1)
    offset: float = attribute(name="offset", default_value=0)


@dataclass
class Unit(SchemaBase):
    """Container for unit information for a series."""

    label: str = attribute(name="label")
    quantity: str = attribute(name="quantity")
    si_unit: List[SIUnit] = element(name="SIUnit", default=list)

    def add_si_unit(self, si_unit: SIUnit) -> None:
        """
        Adds a SI base unit to the container. Must be of SIUnit type.

        Args:
            si_unit (SIUnit): The SI base unit.
        """
        self.si_unit.append(si_unit)



from pyaniml.core import AnIMLDocument
from pyaniml.core import ExperimentStep
from pyaniml.core import SampleReference
from pyaniml.core import SampleReferenceSet
from pyaniml.core import Sample
from pyaniml.core import Series
from pyaniml.core import SeriesSet
from pyaniml.core import IndividualValueSet
from pyaniml.core import Parameter
from pyaniml.core import Category
from pyaniml.core import Infrastructure
from pyaniml.core import Method
from pyaniml.core import Device
from pyaniml.core import Author
from pyaniml.core import Software
from pyaniml.core import Result
from pyaniml.core import Unit
from pyaniml.core import SIUnit
from pyaniml.core import DataTypeChoices
from pyaniml.core import Dependencies
from pyaniml.core import DataTypes
from pyaniml.core import Purposes
from pyaniml.core import SIUnits



from dataclasses import dataclass

from pyaniml.utility.utils import SchemaBase, element
from pyaniml.core.sample import Sample, SampleReferenceSet
from pyaniml.core.enums import Purposes


@dataclass
class Infrastructure(SchemaBase):
    """Container for sample references and infrastructure"""

    sample_references: SampleReferenceSet = element(
        name="SampleReferenceSet", default=SampleReferenceSet
    )

    def add_sample_reference(self, sample: Sample, role: str, sample_purpose: Purposes):
        self.sample_references.add_reference(
            sample=sample, role=role, sample_purpose=sample_purpose
        )



from dataclasses import dataclass

from pyaniml.utility import element, SchemaBase
from pyaniml.core.experimentstep import ExperimentStepSet, ExperimentStep
from pyaniml.core.sample import SampleSet, Sample
from pyaniml.core.sets import AuditTrailEntrySet, SignatureSet


@dataclass
class AnIMLDocument(SchemaBase):
    """Container for an AnIML document"""

    class Meta:
        name = "AnIML"
        # namespace = "urn:org:astm:animl:schema:core:draft:0.90"

    sample_set: SampleSet = element(name="SampleSet", default=SampleSet)

    experiment_step_set: ExperimentStepSet = element(
        name="ExperimentStepSet", default=ExperimentStepSet
    )

    audit_trail_entry_set: AuditTrailEntrySet = element(
        name="AuditTrailEntrySet", default=AuditTrailEntrySet
    )

    signature_set: SignatureSet = element(name="SignatureSet", default=SignatureSet)

    def add_sample(self, sample: Sample):
        """Adds a sample to the sample set

        Args:
            sample (Sample): Container describing a sample
        """
        self.sample_set.add_sample(sample)

    def add_experiment_step(self, experiment_step: ExperimentStep):
        """Adds an experiment step to the experiment step set

        Args:
            experiment_step (ExperimentStep): Container describing an experiment step
        """
        self.experiment_step_set.add_experiment_step(experiment_step)



from dataclasses import dataclass
from typing import List, Union

from pyaniml.utility.utils import SchemaBase, elements
from pyaniml.core.series import SeriesSet
from pyaniml.core.parameter import Category


@dataclass
class Result(SchemaBase):
    """Container for experiment results"""

    results: List[object] = elements(
        choices=(
            {"name": "SeriesSet", "type": SeriesSet},
            {"name": "Category", "type": Category},
        ),
        default=list,
    )

    def add_result(self, result: Union[SeriesSet, Category]) -> None:
        """Adds a measurement result to the the container. Must be of any low-level AnIML type.

        Args:
            result (Union[SeriesSet, Category]): The quantitive measurement results.
        """
        self.results.append(result)



from enum import Enum


class DataTypeChoices(Enum):

    """
    This Enum stores the mapping from native Python types to the corresponding
    types in AnIMLs Series element.
    """

    I = "I"
    F = "F"
    S = "S"
    Boolean = "Boolean"

    @classmethod
    def get(cls):
        return (
            {"name": "I", "type": int},
            {"name": "F", "type": float},
            {"name": "S", "type": str},
            {"name": "Boolean", "type": bool},
        )


class DataTypes(Enum):

    """
    This Enum stores all data types that are supported as an attribute in AnIML
    """

    INT32 = "Int32"
    INT64 = "Int64"
    FLOAT32 = "Float32"
    FLOAT64 = "Float64"
    STRING = "String"


class Dependencies(Enum):

    """
    This Enum stores all types of dependencies found in AnIML.
    """

    DEPENDENT = "dependent"
    INDEPENDENT = "independent"


class Purposes(Enum):

    """
    This Enum stores all supported sample purposes in AnIML.
    """

    PRODUCED = "produced"
    CONSUMED = "consumed"


class UserTypes(Enum):

    """
    This Enum stores all supported user types in AnIML.
    """

    HUMAN = "human"
    DEVICE = "device"
    SOFTWARE = "software"


class SIUnits(Enum):

    """
    This Enum stores all supported SI units in AnIML.
    """

    DIMENSIONLESS = "1"
    METER = "m"
    GRAM = "g"
    SECOND = "s"
    AMPERE = "A"
    KELVIN = "K"
    MOL = "mol"
    CANDELA = "cd"



from typing import List, Union
from dataclasses import dataclass
from pyaniml.core.unit import Unit
from pyaniml.utility.utils import SchemaBase, element, attribute, elements
from pyaniml.core.enums import DataTypeChoices, Dependencies, DataTypes


@dataclass
class IndividualValueSet(SchemaBase):
    """Container holding individual values of a measurement"""

    data: List[object] = elements(default=list, choices=DataTypeChoices.get())


@dataclass
class Series(SchemaBase):
    """Container holding a single Series of datapoints"""

    name: str = attribute()
    id: str = attribute(name="seriesID")
    data: IndividualValueSet = element(name="IndividualValueSet")
    unit: Unit = element(name="Unit")
    dtype: DataTypes = attribute(name="SeriesType")
    dependency: Dependencies = attribute()
    plot_scale: str = attribute(name="plotScale")

    def __post_init__(self):
        """Used to convert a list of values to an IndividualValueSet"""

        self.data = IndividualValueSet(data=self.data)


@dataclass
class SeriesSet(SchemaBase):
    """Container for multi-dimensional datasets"""

    name: str = attribute()
    series: List[Series] = element(name="Series", default=list)



from typing import Any, Dict, Tuple, Optional
from dataclasses import field
from pydantic.json import pydantic_encoder

from xsdata.formats.dataclass.serializers import XmlSerializer, JsonSerializer
from xsdata.formats.dataclass.serializers.config import SerializerConfig
from xsdata.formats.dataclass.context import XmlContext
from xsdata.formats.dataclass.parsers import XmlParser, JsonParser


class SchemaBase:
    def toXML(self) -> str:
        """Serializes an abstract data model to XML.

        Returns:
            str: XML data model as string.
        """
        config = SerializerConfig(pretty_print=True)
        serializer = XmlSerializer(config=config)
        return serializer.render(self)

    def toJSON(self) -> str:
        """Serializes an abstract data model to JSON.

        Args:
            indent (int): Degree of indentation.

        Returns:
            str: JSON data model as string.
        """
        config = SerializerConfig(pretty_print=True)
        serializer = JsonSerializer(context=XmlContext(), config=config)
        return serializer.render(self)

    def toDict(self) -> str:
        """Serializes an abstract data model to a dictionary.

        Returns:
            Dict[Str, Any]: JSON data model as string.
        """
        return pydantic_encoder(self)

    def JSONSchema(self) -> Dict[str, Any]:
        return self.__pydantic_model__.schema()

    @classmethod
    def fromXMLString(cls, xml_string: str):
        """Initializes an object model from an XML string.

        Args:
            xml_string (str): XML data model as string.

        Returns:
            Any: Object model that has been parsed from the XML string.
        """
        parser = XmlParser(context=XmlContext())
        return parser.from_string(xml_string, cls)

    @classmethod
    def fromJSONString(cls, json_string: str):
        """Initializes an object model from a JSON string.

        Args:
            json_string (str): JSON data model as string.

        Returns:
            Any: Object model that has been parsed from the JSON string.
        """

        parser = JsonParser(context=XmlContext())
        return parser.from_string(json_string, cls)


def _generateField(
    type: str,
    name: Optional[str] = None,
    default: Any = None,
    default_value: Any = None,
    choices: Optional[Tuple[Dict[str, Any], ...]] = None,
):
    """Metafunction to generate a valid field for later XML serialization."""

    # Declare dict type
    metadata_dict: Dict[str, Any] = dict(type=type)

    param_dict: Dict[str, Any] = {}

    if choices:
        metadata_dict["choices"] = choices
    if name:
        metadata_dict["name"] = name

    if default:
        param_dict["default_factory"] = default
    if default_value is not None:
        param_dict["default"] = default_value

    return field(**param_dict, metadata=metadata_dict)


def attribute(
    name: Optional[str] = None, default: Any = None, default_value: Any = None
):
    """Defines an attribute as an XML attribute.

        <parent name=XYZ />

    Args:
        name (str): The name of the attribute
        default (Any, optional): Default callable for mutable data types. Defaults to None.

    Returns:
        field: Dataclass Field element, which adds all teh metadata needed to generate an XML model
    """
    return _generateField(
        type="Attribute", name=name, default=default, default_value=default_value
    )


def element(name: Optional[str] = None, default: Any = None, default_value: Any = None):
    """Defines an attribute as an XML element.

        <name>XYZ</name>

    Args:
        name (str): The name of the attribute
        default (Any, optional): Default callable for mutable data types. Defaults to None.

    Returns:
        field: Dataclass Field element, which adds all teh metadata needed to generate an XML model
    """
    return _generateField(
        type="Element", name=name, default=default, default_value=default_value
    )


def elements(choices: Tuple[Dict[str, Any], ...], default: Any = None):
    """Defines an xsd:choices reference as an XML element.

        <choices1>XYZ</choices1>
        <choices2>XYZ</choices2>

    Args:
        choices (str): The types of elements that can be defined in the model.
        default (Any, optional): Default callable for mutable data types. Defaults to None.

    Returns:
        field: Dataclass Field element, which adds all teh metadata needed to generate an XML model
    """
    return _generateField(type="Elements", choices=choices, default=default)



from dataclasses import dataclass
from typing import List, Union

from pyaniml.utility.utils import SchemaBase, attribute, element, elements
from pyaniml.core.parameter import Category
from pyaniml.core.enums import UserTypes


@dataclass
class Software(SchemaBase):
    """Container of software used in this ExerimentStep"""

    manufacturer: str = element("Manufacturer")
    name: str = element("Name")
    version: str = element("Version")
    operating_system: str = element("OperatingSystem")


@dataclass
class Device(SchemaBase):
    """Container describing a physical device"""

    device_id: str = element("DeviceIdentifier")
    manufacturer: str = element("Manufacturer")
    name: str = element("Name")
    firmware_version: str = element("FirmwareVersion")
    serial_number: str = element("SerialNumber")


@dataclass
class Author(SchemaBase):
    """Container describing an author"""

    user_type: UserTypes = attribute("userType")
    name: str = element("Name")
    affiliation: str = element("Affiliation")
    role: str = element("Role")
    email: str = element("Email")
    phone: str = element("Phone")
    location: str = element("Location")


@dataclass
class Method(SchemaBase):
    """Container for method-related informations"""

    methods: List[object] = elements(
        choices=(
            {"name": "Author", "type": Author},
            {"name": "Device", "type": Device},
            {"name": "Software", "type": Software},
            {"name": "Category", "type": Category},
        ),
        default=list,
    )

    def add_method(self, method: Union[Author, Device, Software, Category]):
        """Adds a method-related property to the Method section of an experiment step

        Args:
            method (Union[Author, Device, Software, Category]): Characteristics of a method.
        """
        self.methods.append(method)



import setuptools
from setuptools import setup

setup(
    name="pyAnIML",
    version="0.0.3",
    description="Handling of AnIML files",
    url="https://github.com/FAIRChemistry/pyAnIML",
    author="Range, Jan",
    author_email="jan.range@simtech.uni-stuttgart.de",
    license="MIT License",
    packages=setuptools.find_packages(),
    install_requires=["xsdata", "pydantic"],
    extras_require={"test": ["pytest-cov"]},
)



from typing import List
from dataclasses import dataclass
from pyaniml.utility.utils import SchemaBase, element


@dataclass
class AuditTrailEntrySet(SchemaBase):
    """Container for audit trail entries"""

    audit_trail_entries: List[str] = element(name="AuditTrailEntry", default=list)

    def add_audit_trail_entry(self, audit_trail_entry):
        self.audit_trail_entries.append(audit_trail_entry)


@dataclass
class SignatureSet(SchemaBase):
    """Container for digital signatures"""

    signatures: List[str] = element(name="Signature", default=list)

    def add_signature(self, signature):
        self.signatures.append(signature)



from typing import Union, List
from dataclasses import dataclass

from pyaniml.utility import attribute, element, SchemaBase
from pyaniml.core.sample import Sample
from pyaniml.core.infrastructure import Infrastructure
from pyaniml.core.result import Result
from pyaniml.core.method import Method, Software
from pyaniml.core.method import Device
from pyaniml.core.method import Author
from pyaniml.core.parameter import Category
from pyaniml.core.series import SeriesSet
from pyaniml.core.enums import Purposes


@dataclass
class ExperimentStep(SchemaBase):
    """Container for metadata as well as measured data from an experiment"""

    name: str = attribute()
    experiment_step_id: str = attribute(name="experimentStepID")
    infrastructure: Infrastructure = element(
        name="Infrastructure", default=Infrastructure
    )
    method: Method = element(name="Method", default=Method)
    result: Result = element(name="Result", default=Result)

    def add_sample_reference(self, sample: Sample, role: str, sample_purpose: Purposes):
        """Adds a sample to the infrastructure and creates a Reference.

        Args:
            sample (Sample): Sample object that has been defined beforehand
            role (str): The role of the sample in this experiment step.
            sample_purpose (str): The purpose the sample is intended for.
        """
        self.infrastructure.add_sample_reference(
            sample=sample, role=role, sample_purpose=sample_purpose
        )

    def add_method(self, method: Union[Author, Device, Software, Category]) -> None:
        """Adds a method-related property to the Method section of an experiment step

        Args:
            method (Union[Author, Device, Software, Category]): Characteristics of a method.
        """
        self.method.add_method(method)

    def add_result(self, result: Union[SeriesSet, Category]) -> None:
        """Adds a result to the Result section of an experiment step

        Args:
            result (Union[SeriesSet, Category]): Characteristics of a result.
        """
        self.result.add_result(result)


@dataclass
class ExperimentStepSet(SchemaBase):
    """Container for experiment steps"""

    experiment_steps: List[ExperimentStep] = element(
        name="ExperimentStep", default=list
    )

    def add_experiment_step(self, experiment_step: ExperimentStep) -> None:
        """Adds an experiment step to the experiment set.

        Args:
            experiment_step (ExperimentStep): Object describing an experiment step.
        """
        self.experiment_steps.append(experiment_step)



from dataclasses import dataclass
from typing import List, Union

from pyaniml.utility.utils import attribute, elements, element, SchemaBase
from pyaniml.core.parameter import Parameter, Category
from pyaniml.core.enums import Purposes


@dataclass
class Sample(SchemaBase):
    """Describes an experiment sample"""

    name: str = attribute(name="name")
    id: str = attribute(name="sampleID")
    properties: List[object] = elements(
        choices=(
            {"name": "Parameter", "type": Parameter},
            {"name": "Category", "type": Category},
        ),
        default=list,
    )

    def add_property(self, property: Union[Parameter, Category]) -> None:
        """Adds a property of type [Parameter, Category] to an AnIML document.

        Args:
            property (object): Some kind of property that describes the sample.
        """

        self.properties.append(property)


@dataclass
class SampleReference(SchemaBase):
    """References an experiment sample"""

    sample_id: str = attribute(name="sampleID")
    role: str = attribute(name="role")
    sample_purpose: Purposes = attribute(name="samplePurpose")

    @classmethod
    def from_sample(cls, sample: Sample, role: str, sample_purpose: Purposes):
        return cls(sample_id=sample.id, role=role, sample_purpose=sample_purpose)


@dataclass
class SampleSet(SchemaBase):
    """Container for samples"""

    samples: List[Sample] = element(name="Sample", default=list)

    def add_sample(self, sample: Sample) -> None:
        """Adds a sample to a sample set.

        Args:
            sample (Sample): Object describing a sample.
        """
        self.samples.append(sample)


@dataclass
class SampleReferenceSet(SchemaBase):
    """Container for references to samples that have been defined in the AnIMLDocument"""

    sample_references: List[SampleReference] = element(
        name="SampleReference", default=list
    )

    def add_reference(self, sample: Sample, role: str, sample_purpose: Purposes):
        """Adds a sample reference to a sample set.

        Args:
            sample (Sample): Sample object to ensure correct ID assigment
            role (str): The role of the sample.
            sample_purpose (str): The purpose of the sample.
        """
        self.sample_references.append(
            SampleReference.from_sample(
                sample=sample, role=role, sample_purpose=sample_purpose
            )
        )



from typing import List, Optional, Union
from dataclasses import dataclass

from pyaniml.core.series import SeriesSet
from pyaniml.core.enums import DataTypes
from pyaniml.utility.utils import SchemaBase, attribute, elements


@dataclass
class Parameter(SchemaBase):
    """Container holding a parameter description"""

    value: Union[str, float, int, bool]
    name: str = attribute()
    dtype: DataTypes = attribute(name="parameterType")


@dataclass
class Category(SchemaBase):
    """Container holding a category description with multiple parameters"""

    name: str = attribute()
    content: List[object] = elements(
        choices=(
            {"name": "Parameter", "type": Parameter},
            {"name": "SeriesSet", "type": SeriesSet},
            {"name": "Category", "type": Optional["Category"]},
        ),
        default=list,
    )

    def __post_init__(self):
        # Validate content types
        if not all(isinstance(entry, (Parameter, SeriesSet)) for entry in self.content):
            # Guard clause to check type conistency
            raise TypeError("Property must be either a 'Parameter' or a 'SeriesSet'")

    def add_content(self, content: Union[Parameter, SeriesSet]) -> None:
        """Adds a category to the the container. Must be of any low-level AnIML type.

        Args:
            content (Union[Parameter, SeriesSet]): Content to add to the container.
        """

        self.content.append(content)



from pyaniml.core.animl import AnIMLDocument
from pyaniml.core.experimentstep import ExperimentStep
from pyaniml.core.sample import SampleReference
from pyaniml.core.sample import SampleReferenceSet
from pyaniml.core.sample import Sample
from pyaniml.core.series import Series
from pyaniml.core.series import IndividualValueSet
from pyaniml.core.series import SeriesSet
from pyaniml.core.parameter import Parameter
from pyaniml.core.parameter import Category
from pyaniml.core.infrastructure import Infrastructure
from pyaniml.core.method import Method
from pyaniml.core.method import Device
from pyaniml.core.method import Author
from pyaniml.core.method import Software
from pyaniml.core.result import Result
from pyaniml.core.unit import Unit
from pyaniml.core.unit import SIUnit
from pyaniml.core.enums import DataTypeChoices
from pyaniml.core.enums import Dependencies
from pyaniml.core.enums import DataTypes
from pyaniml.core.enums import Purposes
from pyaniml.core.enums import SIUnits
